﻿using Vintagestory.API.Common;
using System.IO;
using System;
using Vintagestory.API.MathTools;
using System.Collections.Generic;
using Vintagestory.API.Server;
using System.Security.Cryptography.X509Certificates;
using System.Text.Json;
using static System.Runtime.InteropServices.JavaScript.JSType;
using System.Diagnostics;
using Vintagestory.API.Util;
using System.Collections;
using static System.Net.Mime.MediaTypeNames;
using Vintagestory.API.Common.Entities;

namespace grieflogger
{
    static class ListExtension
    {
        public static T PopAt<T>(this List<T> list, int index)
        {
            T r = list[index];
            list.RemoveAt(index);
            return r;
        }
    }

    public class griefloggerModSystem : ModSystem
    {
        const int BROKEN_LINE_LIMIT = 10000;
        public static ICoreAPI api;
        const bool DEBUG = false;
        const string GRIEF_KEY = "GRIEF_KEY";
        private ICoreServerAPI sapi;
        List<string> brokenData;
        Dictionary<string, List<string>>  placedData;

        public override void Start(ICoreAPI api)
        {
            griefloggerModSystem.api = api;
        }

        public override void StartServerSide(ICoreServerAPI api)
        {
            sapi = api;
            api.Event.BreakBlock += Event_BreakBlock;
            api.Event.DidPlaceBlock += Event_DidPlaceBlock;
            api.Event.DidUseBlock += Event_DidUseBlock;
            brokenData = GetData("brokenData") ?? new List<string>();
            //brokenData = new List<string>(); //reset data
            placedData = new Dictionary<string, List<string>>();
            OutputLog(brokenData, "grief_broken_log.txt");
            api.ChatCommands.Create("grieflog")
                .WithDescription("Dump the broken log")
                .RequiresPrivilege(Privilege.kick)
                //.WithArgs(api.ChatCommands.Parsers.Word("cmd", new string[] { "list", "join", "leave" }))
                .HandleWith(new OnCommandDelegate(OnLfg));
        }

        private List<string> GetData(string key)
        {
            byte[] data = sapi.WorldManager.SaveGame.GetData(key);
            if (data == null) return null;
            return SerializerUtil.Deserialize<List<string>>(data);
        }
        private TextCommandResult OnLfg(TextCommandCallingArgs args)
        {
            OutputLog(brokenData, "grief_log.txt");
            return TextCommandResult.Success("Dumped the log successfully!");
        }
        private List<string> ParseLine(string line)
        {
            return new List<string>(line.Split(';'));
        }

        private List<string> CheckForPlacedBlock(string locString, string uid, string chunkID)
        {
            placedData[chunkID] = GetData("placedData" + chunkID) ?? new List<string>();
            foreach (var line in placedData[chunkID])
            {
                List<string> parsed = ParseLine(line);
                //matches location but not uid
                bool locMatch = parsed[1] == locString;
                bool ownerMisMatch = parsed[2] != uid;
                if (DEBUG) ownerMisMatch = true;
                if (locMatch && ownerMisMatch) return parsed;
            }
            return null;
        }
        private void OutputLog(List<string> dataStore, string fn)
        {
            string filePath = Environment.SpecialFolder.ApplicationData.ToString();
            //api.Logger.Debug(filePath);
            string docPath = Path.GetDirectoryName(filePath);
            bool append = false;

            using (StreamWriter outputFile = new StreamWriter(Path.Combine(docPath, fn), append))
            {
                string header = "TIMESTAMP;LOCATION;BREAKER ID;BREAKER NAME;BLOCK ID;BLOCK CODE;OWNER TIMESTAMP;OWNER ID;OWNER NAME";
                outputFile.WriteLine(header);
                foreach (string s in dataStore)
                {
                    outputFile.WriteLine(s);
                }
            }
            return;
        }

        private void Event_BreakBlock(IServerPlayer byPlayer, BlockSelection blockSel, ref float dropQuantityMultiplier, ref EnumHandling handling)
        {
            IWorldChunk chunk = api.World.BlockAccessor.GetChunkAtBlockPos(blockSel.Position);
            string chunkID = chunk.GetModdata<string>(GRIEF_KEY);
            if (chunkID == null) return;
            string locString = blockSel.Position.ToString();
            List<string> oldData = CheckForPlacedBlock(locString, byPlayer.PlayerUID, chunkID);
            if (oldData == null) return;
            string s = DateTime.Now.ToString();         //0
            s += ";" + locString;                       //1
            s += ";" + byPlayer.PlayerUID;              //2
            s += ";" + byPlayer.PlayerName;             //3
            s += ";" + blockSel.Block.Id;               //4
            s += ";" + blockSel.Block.Code.ToString();  //5
            s += ";" + oldData[0];                      //6 original date
            s += ";" + oldData[2];                      //7 original player id
            s += ";" + oldData[3] + ";";                //8 original player name
            brokenData.Add(s);
            while (brokenData.Count > BROKEN_LINE_LIMIT)
            {
                brokenData.PopAt(brokenData.Count - 1);
            }
            sapi.WorldManager.SaveGame.StoreData("brokenData", SerializerUtil.Serialize(brokenData));
        }
        private void Event_DidPlaceBlock(IServerPlayer byPlayer, int oldblockId, BlockSelection blockSel, ItemStack withItemStack)
        {
            
            IWorldChunk chunk = api.World.BlockAccessor.GetChunkAtBlockPos(blockSel.Position);
            string chunkID = chunk.GetModdata<string>(GRIEF_KEY);
            if (chunkID == null)
            {
                Guid guid = Guid.NewGuid();
                chunkID = guid.ToString();
                chunk.SetModdata<string>(GRIEF_KEY, chunkID);
            }
            placedData[chunkID] = GetData("placedData" + chunkID) ?? new List<string>();
            string s = DateTime.Now.ToString();         //0
            s += ";" + blockSel.Position.ToString();    //1
            s += ";" + byPlayer.PlayerUID;              //2
            s += ";" + byPlayer.PlayerName + ";";       //3
            placedData[chunkID].Add(s);
            sapi.WorldManager.SaveGame.StoreData("placedData" + chunkID, SerializerUtil.Serialize(placedData[chunkID]));
        }
        private void Event_DidUseBlock(IServerPlayer byPlayer, BlockSelection blockSel)
        {
            IWorldChunk chunk = api.World.BlockAccessor.GetChunkAtBlockPos(blockSel.Position);
            string chunkID = chunk.GetModdata<string>(GRIEF_KEY);
            if (chunkID == null) return;
            api.Logger.Event("USING BLOCK!");
            string locString = blockSel.Position.ToString();
            List<string> oldData = CheckForPlacedBlock(locString, byPlayer.PlayerUID, chunkID);
            Block block = api.World.BlockAccessor.GetBlock(blockSel.Position);
            if (oldData == null) return;
            string s = DateTime.Now.ToString();         //0
            s += ";" + locString;                       //1
            s += ";" + byPlayer.PlayerUID;              //2
            s += ";" + byPlayer.PlayerName;             //3
            s += ";" + block.Id;                        //4
            s += ";" + block.Code.ToString();           //5
            s += ";" + oldData[0];                      //6 original date
            s += ";" + oldData[2];                      //7 original player id
            s += ";" + oldData[3] + ";";                //8 original player name
            brokenData.Add(s);
            while (brokenData.Count > BROKEN_LINE_LIMIT)
            {
                brokenData.PopAt(brokenData.Count - 1);
            }
            sapi.WorldManager.SaveGame.StoreData("brokenData", SerializerUtil.Serialize(brokenData));
        }
    }
}
